#include <math.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>

#define frag_size 2048 /* bytes   */
#define samp_size 1024 /* samples */
#define period 300

volatile int idevice=0,odevice=0,diskwriter=0;
int arg,status,caps;
pthread_t reader_id, writer_id, processor_id;
pthread_attr_t attr;
struct timeval tv;
fd_set rfds;
unsigned char ichar;
char *ipath=NULL, *opath=NULL, *dpath=NULL;
pthread_mutex_t imutex,omutex;

struct sbuffer  {       volatile int lock; 
                        volatile struct sbuffer *next;
                        volatile signed short int data[samp_size];
                };

volatile struct sbuffer *outroot=NULL, *outend=NULL, *inroot=NULL, *inend=NULL;

volatile int SLOOP_SHUTDOWN=0, SLOOP_EXIT=0, SLOOP_DISK=0, SLOOP_IMEM=0, SLOOP_OMEM=0, SLOOP_FULLDUPLEX=1, SLOOP_SKIP=0, SLOOP_TS=129;





int duplex_open(char *device)
{
        int dev;

        dev = open (device, O_RDWR); /* O_SYNC  */
        if (dev < 0)                    { printf("Cannot open %s.\n",device); return 0; }

        ioctl(dev, SNDCTL_DSP_SETDUPLEX, 0);
        ioctl(dev, SNDCTL_DSP_GETCAPS  , &caps);
        if (!(caps&DSP_CAP_DUPLEX))     { printf("Not fullduplex capable : %s.\n",device); close(dev); return 0; }

        printf("Opened %s.\nFullduplex: YES.\n",device);

        return dev;
}





void close_all(void)
{
        if(idevice==odevice)
        {
                ioctl(idevice, SNDCTL_DSP_RESET, 0);
                if (close(idevice))     printf("Closing audio device failed.\n");
                else                    printf("Closing audio device.\n");
        }
        else
        {
                ioctl(idevice, SNDCTL_DSP_RESET, 0);
                ioctl(odevice, SNDCTL_DSP_RESET, 0);
                if (close(idevice))     printf("Closing input audio device failed.\n");
                else                    printf("Closing input audio device.\n");
                if (close(odevice))     printf("Closing output audio device failed.\n");
                else                    printf("Closing output audio device.\n");
        }

        if (SLOOP_DISK)
        {       if (close(diskwriter))  printf("Closing output file failed.\n");
                else                    printf("Closing output file.\n");
        }
}





struct sbuffer *switch_frag ( struct sbuffer *ptr )
{	
        struct sbuffer *help;
	int x;

	ptr->next=NULL;
	pthread_mutex_lock(&omutex);
    	    if(outend)	{ outend->next=ptr; outend=ptr; SLOOP_OMEM++; }
	    else     	{ outend=outroot=ptr; 		SLOOP_OMEM=1; }
	
	if (SLOOP_OMEM>86) 
	{   help=(struct sbuffer *)outroot;
	    for(x=0;x<86;x++) { help->lock=x; help=(struct sbuffer *)help->next; }
	}
	
    	pthread_mutex_unlock(&omutex);	
	
    	ptr=(struct sbuffer * ) malloc(sizeof(struct sbuffer));
    	if(ptr==NULL) { SLOOP_SHUTDOWN=2; printf("OUT OF SLOOP_PROCESSOR MEMORY\n"); pthread_exit(NULL); }
	
        return(ptr);
}





void * reader (void * dummy)
{
        struct sbuffer *ptr;
        int st;

        while (SLOOP_SHUTDOWN==0)
        {
            
                ptr=(struct sbuffer * ) malloc(sizeof(struct sbuffer));
                if(ptr==NULL)		{ SLOOP_SHUTDOWN=2; printf("OUT OF SLOOP_READER MEMORY\n"); pthread_exit(NULL); }
                st=read(idevice, (void *)&(ptr->data[0]), frag_size);
                if(st!=frag_size) 	{ SLOOP_SHUTDOWN=2; printf("READER ERROR\n"); pthread_exit(NULL); }

                ptr->next=NULL;

                pthread_mutex_lock(&imutex);
            	    if (inend==NULL)	{ inend=inroot=ptr;            SLOOP_IMEM=1; }
            	    else		{ inend->next=ptr; inend=ptr;  SLOOP_IMEM++; }
                pthread_mutex_unlock(&imutex);
            
        }

        if (!(--SLOOP_SHUTDOWN)) SLOOP_EXIT=1;
        pthread_exit(NULL);
}





void * processor (void * dummy)
{
        struct sbuffer *ptr;
        long int a,c,x,y,z,n,temp,temp2;
        long int multi,multx[period/2],ctrl;

        int found[samp_size*2][3];
        int workptr=0,fragptr=0,walker=0;
        signed short int worker[samp_size*3],wrapper[samp_size];

        temp=0; z=0; multi=0; ctrl=0;

        while (SLOOP_SHUTDOWN==0)
        {
        
	                pthread_mutex_lock(&imutex);

                if(inroot)
                {
                        ptr=(struct sbuffer *)inroot;
                        if (inroot->next) inroot=inroot->next;
                        else inend=inroot=NULL;
			SLOOP_IMEM--;

                        pthread_mutex_unlock(&imutex);

                        memcpy((void *)&(worker[workptr]), (void *)&(ptr->data[0]), frag_size);
                        workptr+=samp_size; found[0][0]=0; n=0; z=0;
                        if (fragptr)
                                memcpy( (void *)&(ptr->data[0]) , (void *)&wrapper[0] , fragptr*2);

                        for(x=1;x<(workptr-1);x++)
                        {
                                if (    ( (worker[x-1]>0) && (worker[x-1]>worker[x]) && (worker[x]>worker[x+1]) && (worker[x+1]<0) )||
                                        ( (worker[x-1]<0) && (worker[x-1]<worker[x]) && (worker[x]<worker[x+1]) && (worker[x+1]>0) )
                                   )
                                {
                                        switch (n)
                                        {
                                                case 2:
                                                {       if ( ( ((x-found[z][1])<=441) && ((x-found[z][1])>=6) ) && ( (x-found[z][1])>=((found[z][1]-found[z][0])/8) ) )
                                                        {       temp2=0; for (y=found[z][1]; y<x; y++) temp2=temp2+worker[y];
                                                                if ( ( abs(temp2/(x-found[z][1])) > 512 ) && ( (temp2/abs(temp2))!=(temp/abs(temp)) )  )
                                                                { found[z][n]=x; n++; break; }
                                                                else
                                                                { n=1; }
                                                        }
                                                        else    { n=1; }
                                                }

                                                case 1:
                                                {       if ( ((x-found[z][0])<=441) && ((x-found[z][0])>=6) )
                                                        {       temp=0; for (y=found[z][0]; y<x; y++) temp=temp+worker[y];
                                                                if ( abs(temp/(x-found[z][0])) > 512 )
                                                                { found[z][n]=x; n++; break; }
                                                                else
                                                                { n=0; }
                                                        }
                                                        else    { n=0; }
                                                }

                                                case 0:
                                                {       found[z][n]=x;
                                                        n++;
                                                        break;
                                                }

                                                default:
                                                        break;
                                        }

                                        if (n==3)
                                        {
                                                found[z+1][0]=found[z][2]; z++; n=1;
                                        }
                                }
                        }

                        walker=0;
			
                        if(0)   /* pass through for testing, set to one */
			{
				    memcpy( (void *)&(ptr->data[fragptr]) , (void *)&worker[0] , (samp_size-fragptr)*2 );
                            	    ptr=switch_frag(ptr); walker=samp_size-fragptr; fragptr=0;
			    
			}			
			else if (z==0)
                        {       
				if(SLOOP_SKIP<SLOOP_TS)
				{    
				    memcpy( (void *)&(ptr->data[fragptr]) , (void *)&worker[0] , (samp_size-fragptr)*2 );
                            	    ptr=switch_frag(ptr); walker=samp_size-fragptr; fragptr=0;
				}
				else if (SLOOP_SKIP==SLOOP_TS)
				{
				    memcpy( (void *)&(ptr->data[0]) , (void *)&worker[0] , frag_size );
				    for(x=0;x<samp_size;x++) ptr->data[x]=((ptr->data[x])*(samp_size-x))/samp_size;				    
                            	    ptr=switch_frag(ptr); walker=samp_size; fragptr=0;
				}				
				else
				{
				    free((void *)ptr); ptr=NULL; 
				    walker=samp_size; fragptr=0;
				}	
	    			
				if (SLOOP_SKIP<=SLOOP_TS) SLOOP_SKIP++;
                        }
			else if (SLOOP_SKIP>SLOOP_TS)
			{
				SLOOP_SKIP=0; ctrl=0; multi=0;
				    memcpy( (void *)&(ptr->data[0]) , (void *)&worker[0] , frag_size );
				    for(x=0;x<samp_size;x++) ptr->data[x]=((ptr->data[x])*x)/samp_size;				    
                            	    ptr=switch_frag(ptr); walker=samp_size; fragptr=0;
			}
                        else
                        {	
				SLOOP_SKIP=0;
			
                                for(x=0;x<z;x++)
                                {

                                ctrl++; ctrl%=period;
        			if (ctrl<(period/2)) { multi+= z/2 - z/16 - z/32 + (multi*multi)/(5000*z); multx[ctrl]=multi;  } 
                                else                 { multi=multx[period-1-ctrl];   }
                                if (multi<0) multi=0;		
				else if (multi>100000) { multx[ctrl]=multi=100000; ctrl=period-1-ctrl; }
				
					c=(found[x][0]-walker);
                                        if (c>0)
                                        {
                                                while (c>(y=samp_size-fragptr))
                                                {
                                                        memcpy( (void *)&(ptr->data[fragptr]) , (void *)&worker[walker]  , y*2 );
                                                        ptr=switch_frag(ptr);
                                                        fragptr=0; c-=y; walker+=y;
                                                }

                                                memcpy( (void *)&(ptr->data[fragptr]) , (void *)&worker[walker]  , c*2 );
                                                fragptr+=c; walker+=c;
                                                if (fragptr==samp_size) { ptr=switch_frag(ptr); fragptr=0; }
                                        }
	
                                        c=(found[x][2]-found[x][0]);
                                        for(a=0;a<=multi/c;a++)
                                        {
                                                y=samp_size-fragptr;
                                                if(c>y)
                                                {
                                                        memcpy( (void *)&(ptr->data[fragptr]) , (void *)&worker[found[x][0]]   , (y)*2 );
                                                        ptr=switch_frag(ptr);
                                                        memcpy( (void *)&(ptr->data[0]) ,       (void *)&worker[found[x][0]+y] , (c-y)*2  );
                                                        fragptr=(c-y);
                                                }
                                                else
                                                {
                                                        memcpy( (void *)&(ptr->data[fragptr]) , (void *)&worker[found[x][0]]   , c*2 );
                                                        fragptr+=c;
                                                        if (fragptr==samp_size) { ptr=switch_frag(ptr); fragptr=0; }
                                                }
                                        }
                                        walker=found[x][2];
                                }
                        }

                        x=workptr-walker;
                        if (x>0)
                        {
                                memmove( (void *)&worker[0] , (void *)&worker[walker], x*2 );
                                workptr=x;
                        }
                        else    workptr=0;

                        if (ptr)
			{ 
			    if(fragptr) { memcpy( (void *)&wrapper[0] , (void *)&(ptr->data[0]) , fragptr*2); }
			    free((void *)ptr); 
			}

                }
                else
                {       pthread_mutex_unlock(&imutex);  }

        }

        if (!(--SLOOP_SHUTDOWN)) SLOOP_EXIT=1;
        pthread_exit(NULL);
}





void * writer (void * dummy)
{
        struct sbuffer *ptr;
        int st,disk_counter=0,run=0;

        while (SLOOP_SHUTDOWN==0)
        {
                        pthread_mutex_lock(&omutex);
				                
			if      ( (run==0) && (SLOOP_OMEM> 4) ) run=1;
		        else if ( (run!=0) && (SLOOP_OMEM==0) ) run=0;        
	
                if (run)
                {
                        
                                ptr=(struct sbuffer *)outroot;
                                if (outroot->next)  outroot=outroot->next;
                                else 		    outend=outroot=NULL; 
			        SLOOP_OMEM--;		
                        pthread_mutex_unlock(&omutex);

                        st=write(odevice, (void *)&(ptr->data[0]), frag_size);
                        if(st!=frag_size) { SLOOP_SHUTDOWN=2; printf("WRITER ERROR\n"); pthread_exit(NULL); }

                        if (SLOOP_DISK)
                        {       st=write(diskwriter, (void *)&(ptr->data[0]), frag_size);
                                if(st!=frag_size)
                                {       printf("ERROR while writing output file. Closing.\n");
                                        if (close(diskwriter)) printf("Closing output file failed.\n");
                                        SLOOP_DISK=0;
                                }
                                else
                                {       disk_counter++;
                                        if (disk_counter>=1048576)
                                        {       printf("Maximum file size reached.\n");
                                                if (close(diskwriter))  printf("Closing output file failed.\n");
                                                else                    printf("Closing output file.\n");
                                                SLOOP_DISK=0;
                                        }
                                }
                        }

                        free((void *)ptr);
                }
		else
		{
                        pthread_mutex_unlock(&omutex);		
		}
        }

        if (!(--SLOOP_SHUTDOWN)) SLOOP_EXIT=1;
        pthread_exit(NULL);
}





int main (int argc, char** argv)
{
        struct sbuffer *ptr, *help;
        int x,y,z;

        char roto[] = "\\|/-";
        char *ptxt[]  = {       "      [recording]\n",
                                "      [skipping ]\n"
                        };
			

        printf("\nsloop.splitter\n\n");

        if (argc>1)
        {
                x=1;
                while (x<argc)
                {
                        if (!strcmp(argv[x],"-of"))
                        {       x++;
                                if (x<(argc))   { dpath=argv[x]; }
                                else            { printf("wrong arguments\n"); exit(1); }
                        }
                        else if (!strcmp(argv[x],"-hd"))
                        {       x++; x++;
                                if (x<(argc))   { ipath=argv[x-1]; opath=argv[x]; }
                                else            { printf("wrong arguments\n"); exit(1); }
                        }
                        else if (!strcmp(argv[x],"-fd"))
                        {       x++;
                                if (x<(argc))   { ipath=argv[x]; }
                                else            { printf("wrong arguments\n"); exit(1); }
                        }
                        else if (!strcmp(argv[x],"-st"))
                        {       x++;
                                if (x<(argc))   { SLOOP_TS=(43*atoi(argv[x]))/1000; }
                                else            { printf("wrong arguments\n"); exit(1); }
                        }			
                        else if (!(strcmp(argv[x],"--help"))||!(strcmp(argv[x],"-h")))
                        {
                                printf("options:\n");
                                printf(" -fd <device>              Fullduplex\n");
                                printf(" -hd <device> <device>     Halfduplex\n");
				printf(" -st <milliseconds>        skipping threshold\n");
                                printf(" -of <path/filename>       optional Output File\n\n");
                                printf("<device> is e.g. /dev/dsp or /dev/dsp4\n\n");
                                exit(0);
                        }
                        else if (!(strcmp(argv[x],"--version"))||!(strcmp(argv[x],"-v")))
                        {
                                printf("version 0.2.1\n\n");
                                exit(0);
                        }
                        else
                        { printf("wrong arguments\n"); exit(1); }

                        x++;
                }
        }

        if (ipath==NULL)
        {
                printf("Assuming fullduplex on /dev/dsp or /dev/dsp0.\n");
                SLOOP_FULLDUPLEX=1;

                idevice=duplex_open("/dev/dsp");
                if(!idevice)
                {       idevice=duplex_open("/dev/dsp0");
                        if(!idevice) exit(1);
                }
        }

        else if (opath==NULL)
        {
                printf("Trying fullduplex on %s.\n",ipath);
                SLOOP_FULLDUPLEX=1;

                idevice=duplex_open(ipath);
                if(!idevice) exit(1);
        }

        else
        {
                printf("Trying halfduplex on %s and %s.\n", ipath, opath);
                SLOOP_FULLDUPLEX=0;
        }


        if(SLOOP_FULLDUPLEX)    /* FULLDUPLEX */
        {
                ioctl(idevice, SNDCTL_DSP_RESET, 0);
		
		arg = 0x7FFF000B;
/*		status = ioctl(idevice, SNDCTL_DSP_SETFRAGMENT, &arg);
		if(status)
		{
			printf("Cannot set fragment size.\n");
			close(idevice);
			exit(1);
		}
*/		
                arg = AFMT_S16_LE; /* sample size */
	        status = ioctl(idevice, SOUND_PCM_WRITE_BITS, &arg);
	        if (status == -1)
                {       printf("SOUND_PCM_WRITE_BITS ioctl failed.\n");
                        close(idevice);
                        exit(1);
                }
	        if (arg != AFMT_S16_LE)
                {       printf("Unable to set sample size.\n");
                        close(idevice);
                        exit(1);
                }

	        arg = 1;  /* mono or stereo */
	        status = ioctl(idevice, SOUND_PCM_WRITE_CHANNELS, &arg);
	        if (status == -1)
                {       printf("SOUND_PCM_WRITE_CHANNELS ioctl failed.\n");
                        close(idevice);
                        exit(1);
                }
	        if (arg != 1)
                {       printf("Unable to set number of channels.\n");
                        close(idevice);
                        exit(1);
                }

                arg = 44100;      /* sampling rate */
	        status = ioctl(idevice, SOUND_PCM_WRITE_RATE, &arg);
	        if (status == -1)
                {       printf("SOUND_PCM_WRITE_WRITE ioctl failed.\n");
                        close(idevice);
                        exit(1);
                }
                odevice=idevice;
        }
        else                    /* HALFDUPLEX */
        {
                idevice = open (ipath, O_RDONLY|O_SYNC);		/* INPUT DEVICE */
                if (idevice < 0)
                {       printf("Open of %s failed.\n",ipath);
                        exit(1);
                }

                ioctl(idevice, SNDCTL_DSP_RESET, 0);

		arg = 0x7FFF000B;
/*		status = ioctl(idevice, SNDCTL_DSP_SETFRAGMENT, &arg);
		if(status)
		{
			printf("Cannot set fragment size.\n");
			close(idevice);
			exit(1);
		}
*/		
                arg = AFMT_S16_LE; /* sample size */
	        status = ioctl(idevice, SOUND_PCM_WRITE_BITS, &arg);
	        if (status == -1)
                {       printf("SOUND_PCM_WRITE_BITS ioctl failed.\n");
                        close(idevice);
                        exit(1);
                }
	        if (arg != AFMT_S16_LE)
                {       printf("Unable to set sample size.\n");
                        close(idevice);
                        exit(1);
                }

	        arg = 1;  /* mono or stereo */
	        status = ioctl(idevice, SOUND_PCM_WRITE_CHANNELS, &arg);
	        if (status == -1)
                {       printf("SOUND_PCM_WRITE_CHANNELS ioctl failed.\n");
                        close(idevice);
                        exit(1);
                }
	        if (arg != 1)
                {       printf("Unable to set number of channels.\n");
                        close(idevice);
                        exit(1);
                }

                arg = 44100;      /* sampling rate */
	        status = ioctl(idevice, SOUND_PCM_WRITE_RATE, &arg);
	        if (status == -1)
                {       printf("SOUND_PCM_WRITE_WRITE ioctl failed.\n");
                        close(idevice);
                        exit(1);
                }

                printf("Opened %s successfully for input.\n",ipath);

                odevice = open (opath, O_WRONLY|O_SYNC);		/* OUTPUT DEVICE */
	        if (odevice < 0)
                {       printf("Open of %s failed.\n",opath);
                        close(idevice);
                        exit(1);
                }

                ioctl(odevice, SNDCTL_DSP_RESET, 0);

		arg = 0x7FFF000B; /* fragment size */
/*		status = ioctl(idevice, SNDCTL_DSP_SETFRAGMENT, &arg);
		if(status)
		{
			printf("Cannot set fragment size.\n");
			close(idevice);
			exit(1);
		}
*/
                arg = AFMT_S16_LE; /* sample size */
	        status = ioctl(odevice, SOUND_PCM_WRITE_BITS, &arg);
	        if (status == -1)
                {       printf("SOUND_PCM_WRITE_BITS ioctl failed.\n");
                        close_all();
                        exit(1);
                }
	        if (arg != AFMT_S16_LE)
                {       printf("Unable to set sample size.\n");
                        close_all();
                        exit(1);
                }

	        arg = 1;  /* mono or stereo */
	        status = ioctl(odevice, SOUND_PCM_WRITE_CHANNELS, &arg);
	        if (status == -1)
                {       printf("SOUND_PCM_WRITE_CHANNELS ioctl failed.\n");
                        close_all();
                        exit(1);
                }
	        if (arg != 1)
                {       printf("Unable to set number of channels.\n");
                        close_all();
                        exit(1);
                }

                arg = 44100;      /* sampling rate */
	        status = ioctl(odevice, SOUND_PCM_WRITE_RATE, &arg);
	        if (status == -1)
                {       printf("SOUND_PCM_WRITE_WRITE ioctl failed.\n");
                        close_all();
                        exit(1);
                }

                printf("Opened %s successfully for output.\n",opath);
        }



        printf("Device parameters set.\n");

        if (pthread_mutex_init(&imutex, NULL) < 0)
        {
                printf("Input mutex failed.\n");
                close_all();
                exit(1);
        }

        if (pthread_mutex_init(&omutex, NULL) < 0)
        {
                printf("Output mutex failed.\n");
                close_all();
                exit(1);
        }

        if(dpath!=NULL)
        {

                diskwriter=open(dpath,O_WRONLY|O_CREAT|O_TRUNC );
                if (diskwriter==-1)     printf("Creating ouput file %s failed.\n",dpath);
                else
                {       SLOOP_DISK=1;
                        printf("Creating ouput file %s\n",dpath);
                        if (fchmod(diskwriter,S_IWUSR|S_IRUSR)) printf("Cannot set file permissions.\n");
                }
        }

	printf("Skipping threshold: %d milliseconds.\n",(SLOOP_TS*1000)/43);

        pthread_attr_init(&attr);
        pthread_create(&reader_id,   &attr, reader,   NULL);
        pthread_create(&writer_id,   &attr, writer,   NULL);
        pthread_create(&processor_id,&attr, processor,NULL);

        printf("Threads started.\nEntering main loop.\n\nHit 'q' <enter> to quit.\n\n\n\n\n\n\n"); z=0;

        do {
                FD_ZERO(&rfds);	
                FD_SET(0, &rfds);		
                tv.tv_sec=0; tv.tv_usec=500000;
                ichar=0;
                if ( select(1, &rfds,NULL,NULL,&tv) ) { ichar=getchar(); }
                pthread_mutex_lock(&imutex); x=SLOOP_IMEM; pthread_mutex_unlock(&imutex);
                pthread_mutex_lock(&omutex); y=SLOOP_OMEM; pthread_mutex_unlock(&omutex);
                if ( ichar!='q' && !SLOOP_EXIT )
                {       printf("\033[5A\rInputQueue :%5d %cByte\nOutputQueue:%5d %cByte\nTotal      :%5d %cByte\n%s\n    \r(%c) [", x>512?x/512:x*2, x>512?'M':'k', y>512?y/512:y*2, y>512?'M':'k', (y+x)>512?(y+x)/512:(y+x)*2, (y+x)>512?'M':'k',(SLOOP_SKIP==SLOOP_TS+1)?ptxt[1]:ptxt[0],roto[(z++)%4]);
                        fflush(stdout);
                }
                else break;
        } while (1);

        printf("\nExiting ... please wait     "); fflush(stdout);

        if(!SLOOP_EXIT)
        {
                SLOOP_SHUTDOWN=3;x=0;
                while(SLOOP_SHUTDOWN)   {       printf("\b\b\b\b(%c) ",roto[(++x)%4]); fflush(stdout);
                                                tv.tv_sec=0; tv.tv_usec=100000;
                                                select(0,NULL,NULL,NULL,&tv);
                                        }
        }

        printf("\n"); fflush(stdout);

        close_all();

        printf("Cleaning up memory     "); fflush(stdout);

        help=(struct sbuffer *)inroot;
        while (help) { ptr=help; help=(struct sbuffer *)help->next; free((void *)ptr); printf("\b\b\b\b(%c) ",roto[(++x)%4]); fflush(stdout); }
        help=(struct sbuffer *)outroot;
        while (help) { ptr=help; help=(struct sbuffer *)help->next; free((void *)ptr); printf("\b\b\b\b(%c) ",roto[(++x)%4]); fflush(stdout); }

        printf("\n");
        return 0;

}
